import assert from "assert";
import { encodeFunctionData } from "viem";
import { Erc20, Erc4337, SmartAccountV1, UniswapV3 } from "./abi";
import { DEX_CHAINS, ERC4337_CHAINS } from "./contracts";
export class SmartAccountV1Provider {
    constructor(publicClient, entrypoint, smartAccountFactory, dex) {
        this.publicClient = publicClient;
        if (dex) {
            this.dex = dex;
        }
        else {
            const dexChain = DEX_CHAINS[publicClient.chain.id];
            if (dexChain) {
                this.dex = dexChain;
            }
        }
        const erc4337Chain = ERC4337_CHAINS[publicClient.chain.id];
        if (entrypoint) {
            this.entrypoint = entrypoint;
        }
        else {
            if (!erc4337Chain || !erc4337Chain.entrypoint) {
                throw new Error("Entrypoint not supported on this chain");
            }
            this.entrypoint = erc4337Chain.entrypoint;
        }
        if (smartAccountFactory) {
            this.smartAccountFactory = smartAccountFactory;
        }
        else {
            if (!erc4337Chain || !erc4337Chain.smartAccountFactoryV1) {
                throw new Error("Smart account V1 not supported on this chain");
            }
            this.smartAccountFactory = erc4337Chain.smartAccountFactoryV1;
        }
    }
    async createSwapRequest(createSwapRequestInput) {
        if (!createSwapRequestInput.smartAccount && !createSwapRequestInput.initSmartAccountInput) {
            throw new Error("Must provide either `smartAccount` or `initSmartAccountInput`");
        }
        if (!this.dex) {
            throw new Error("Dex not supported");
        }
        const dexName = createSwapRequestInput.dex;
        const dex = this.dex[dexName];
        if (!dex) {
            throw new Error(`${dexName} not supported`);
        }
        let smartAccount = createSwapRequestInput.smartAccount;
        let initCode = "0x";
        const initSmartAccountInput = createSwapRequestInput.initSmartAccountInput;
        if (initSmartAccountInput) {
            const { owner, salt } = initSmartAccountInput;
            const address = await this.publicClient.readContract({
                abi: SmartAccountV1.FACTORY_ABI,
                address: this.smartAccountFactory,
                functionName: "getAddress",
                args: [owner, salt],
            });
            const bytecode = await this.publicClient.getBytecode({ address });
            if (bytecode && bytecode !== "0x") {
                throw new Error("Smart account already exists");
            }
            const initCodeCallData = encodeFunctionData({
                abi: SmartAccountV1.FACTORY_ABI,
                functionName: "createAccount",
                args: [owner, salt],
            });
            initCode = (this.smartAccountFactory.toLowerCase() + initCodeCallData.slice(2));
            smartAccount = address.toLowerCase();
        }
        assert(smartAccount !== undefined);
        const { tokenIn, tokenOut, fee, recipient, deadline, amountIn, amountOutMinimum } = createSwapRequestInput.swapInput;
        const approveCallData = encodeFunctionData({
            abi: Erc20.ABI,
            functionName: "approve",
            args: [dex.swapRouter, amountIn],
        });
        const exactInputSingleCallData = encodeFunctionData({
            abi: UniswapV3.SWAP_ROUTER_ABI,
            functionName: "exactInputSingle",
            args: [
                {
                    tokenIn,
                    tokenOut,
                    fee,
                    recipient: recipient || smartAccount,
                    deadline,
                    amountIn,
                    amountOutMinimum,
                    sqrtPriceLimitX96: BigInt(0),
                },
            ],
        });
        const dest = [tokenIn, dex.swapRouter];
        const value = [];
        const func = [approveCallData, exactInputSingleCallData];
        const executeBatchCallData = encodeFunctionData({
            abi: SmartAccountV1.ABI,
            functionName: "executeBatch",
            args: [dest, value, func],
        });
        // nonce is uint256 with 192-bit key and 64-bit value
        // key is pool address
        const nonce = await this.getNonceForSmartAccountV1(smartAccount, dex.factory, [tokenIn, tokenOut], fee);
        // ==================================================
        // ===== maximum of values before is "< uin120" =====
        // ==================================================
        // Note: Step 1 -> 3 is gasLimit, 4 is gasPrice
        // So, (1 + 2 + 3) * 4 = total `wei` needs to execute this transaction
        // 1. calculate callGasLimit => Gas limit for execution phase, estimate by using EntryPoint simulation (function simulateValidation)
        const callGasLimit = await this.calculateExecuteBatchGasLimit(smartAccount, dest, value, func);
        // 2. calculate verificationGasLimit => Gas limit for verification phase, estimate by using EntryPoint simulation (function simulateValidation)
        const verificationGasLimit = await this.calculateVerificationGasLimitV1(initCode !== "0x");
        // 3. calculate preVerificationGas => Gas to compensate the bundler, Use internal bundler => Don't care
        const preVerificationGas = BigInt(0);
        // 4. calculate maxFeePerGas => 30 gwei
        // 5. calculate maxPriorityFeePerGas => estimate: 2 gwei
        // TODO: Both fees based on particular chain
        let maxFeePerGas = BigInt(0);
        let maxPriorityFeePerGas = BigInt(0);
        if (!createSwapRequestInput.gasless) {
            maxFeePerGas = BigInt(30e9);
            maxPriorityFeePerGas = BigInt(2e9);
        }
        const userOpWithoutSign = {
            sender: smartAccount,
            nonce,
            initCode,
            callData: executeBatchCallData,
            callGasLimit,
            verificationGasLimit,
            preVerificationGas,
            maxFeePerGas,
            maxPriorityFeePerGas,
            paymasterAndData: "0x",
            signature: "0x",
        };
        // userOpHash does not depend on signature
        const userOpHash = await this.publicClient.readContract({
            abi: Erc4337.ENTRYPOINT_ABI,
            address: this.entrypoint,
            functionName: "getUserOpHash",
            args: [userOpWithoutSign],
        });
        return { smartAccount, userOpHash, request: userOpWithoutSign };
    }
    async getNonceForSmartAccountV1(smartAccount, dexFactory, tokens, fee) {
        if (tokens.length !== 2) {
            throw new Error("Invalid token pair");
        }
        const poolAddr = await this.publicClient.readContract({
            abi: UniswapV3.FACTORY_ABI,
            address: dexFactory,
            functionName: "getPool",
            args: [tokens[0], tokens[1], fee],
        });
        const formattedPoolAddr = poolAddr.toLowerCase();
        const key = BigInt(formattedPoolAddr);
        const nonce = await this.publicClient.readContract({
            abi: Erc4337.ENTRYPOINT_ABI,
            address: this.entrypoint,
            functionName: "getNonce",
            args: [smartAccount, key],
        });
        return nonce;
    }
    async calculateVerificationGasLimitV1(initSmartAccount) {
        if (initSmartAccount) {
            return BigInt(300000);
        }
        else {
            return BigInt(50000);
        }
    }
    async calculateExecuteBatchGasLimit(smartAccount, dest, value, func) {
        const estimatedGas = await this.publicClient.estimateContractGas({
            abi: SmartAccountV1.ABI,
            address: smartAccount,
            functionName: "executeBatch",
            args: [dest, value, func],
            account: this.entrypoint,
        });
        // ref: https://github.com/wolflo/evm-opcodes/blob/main/gas.md#aa-call-operations
        const baseGas = BigInt(21000);
        // TODO: smart account does not exist, so can not estimate gas
        if (estimatedGas < 30000) {
            return 170000n;
        }
        return estimatedGas - baseGas;
    }
}
//# sourceMappingURL=smartAccountV1Provider.js.map